package algorithms;

import graph.AdjacencyMatrixGraph;
import graph.ListColoredAdjacencyMatrixGraph;
import importer.Results.Instance;

import java.util.*;

public class IterativeSolver extends Solver {
	final double  maxTime = 60000*30;
	double startTime;
	private MatchingType type;
	private final boolean verbose = false;
	private int[] currentMatching; 									//Current matching that algorithm has
	private boolean[] fixedMatches; 								//Keeps track of fixed vertices in current matching
	private int n,positionPointer;
	//private OutWriter out;
	private Instance instance;
	public enum ConflictType {None,Edge,Label,Unresolvable};
	
	private class Conflict {
		int u,v;
		ConflictType type;
		
		public Conflict(ConflictType t){
			this.type = t;
		}
		
		public Conflict(int u, int v, ConflictType t){
			this.u = u;
			this.v = v;
			this.type = t;
		}	
	}
	
	
	public IterativeSolver(ListColoredAdjacencyMatrixGraph g, AdjacencyMatrixGraph h, MatchingType m, Instance instance){
		super(g,h);
		type = m;
		this.currentMatching = new int[g.vertexNum()];
		this.fixedMatches = new boolean[g.vertexNum()];
		//this.out = out;
		this.instance = instance;
		n = g.vertexNum();
	}
	
	@Override
	public boolean solve() {
		startTime = System.currentTimeMillis();
		ReductionRules r = new ReductionRules(this);
		r.reduceInstance();
		return findMatch(0);
	}
	
	/**
	 * Recursive search tree method that finds 
	 * a matching of g in h which respects the list-coloring
	 * 
	 * @param  i  the first non-matched vertex
	 */
	private boolean findMatch(int i){
		if (verbose) System.out.println("i: "+i);
		if (i==g.vertexNum()) {
			for (int j=0;j<currentMatching.length;j++){
				if (verbose) System.out.println(currentMatching[j]);				
			}
			if (verbose) System.out.println("Match found");
			instance.addMatching(getMatchingMap());
			return true;
		}
		else{
			for (int j=0;j<i;j++) fixedMatches[j]=false;
			fixedMatches[i]=true;
			int[] matchCopy = Arrays.copyOf(currentMatching,n);
			boolean[] fixCopy = Arrays.copyOf(fixedMatches,n);
			//try all possibilities for i		
			for (int v :g.getList(i)){
				//regain currentMatching of v
				currentMatching = Arrays.copyOf(matchCopy,n);
				//remove all matching-fixes
				fixedMatches = Arrays.copyOf(fixCopy,n);
				currentMatching[i] = v;				
				Conflict conf = getConflict(i); 
				if (conf.type==ConflictType.None) {
					//new current matching is found
					if (verbose) System.out.println("No problem with next vert");
					return findMatch(i+1);
				}
				else {
					if (verbose) System.out.println("Mhh A Problem can it be resolved?");
					if (resolveMatchingStart(i,conf)){//conflicts can be resolved
						return findMatch(i+1);
					}
				}
			}
		}
		if (verbose) System.out.println("No Match Found");
		return false;
	}
	
	private boolean resolveMatchingStart(int currentPos, Conflict conf) {
		int confVert=-1;
		positionPointer=currentPos;
		if (conf.u==currentPos) confVert = conf.v;
		else if (conf.v==currentPos) confVert = conf.u;
		else if (verbose) {
			System.out.println("Should also not happen... i:"+currentPos);
			System.out.println("Conflict type:" + conf.type);
		}
		return branch(currentPos,confVert);
	}
	
	
	private boolean resolveMatching() {
		if (System.currentTimeMillis()-startTime>maxTime){
			instance.terminate();
			return false;
		}
		/*		
		 * checks for new conflicts in current matching
		 * branches if needed and still possible
		 */
		if (verbose) System.out.println("resolve Matching called"+positionPointer);
		Conflict conf = getConflict(positionPointer);
		if (verbose) System.out.println(conf.u+" "+conf.v);
		if (conf.type==ConflictType.None) {
			for (int j=0;j<currentMatching.length;j++){
				if (verbose) System.out.println(currentMatching[j]);				
			}
			if (verbose) System.out.println("No conflict");
			return true;
		}
		else if (conf.type==ConflictType.Unresolvable) {
			if (verbose) System.out.println("unresolvable");
			return false;
		}
		else {
			if (fixedMatches[conf.u]) return branch(conf.u,conf.v);
			else return branch(conf.v,conf.u);
		}
	}

	private boolean areConflict(int i, int j) {
		int v = currentMatching[i];
		int w = currentMatching[j];
		if (this.type==MatchingType.Homomorph){				
			if (g.areAdjacent(i, j)){//if i and j are adjacent
				if (!h.areAdjacent(w, v)) 
					return true; //then their matching verts in h must be also adjacent
			}
		}
		if (this.type==MatchingType.Isomorph){				
			if (g.areAdjacent(i,j) != h.areAdjacent(w, v)) return true; //then their matching verts in h must be also adjacent								
		}
		if (this.type==MatchingType.All){				
			return false; //no two vertices can form conflict								
		}
		return false;
	}

		
	private boolean branch(int currentPos, int confVert){
		int[] matchCopy = Arrays.copyOf(currentMatching,n);
		boolean[] fixCopy = Arrays.copyOf(fixedMatches,n);		
		for (int v: g.getList(confVert)){
			//regain currentMatching of v
			currentMatching = Arrays.copyOf(matchCopy,n);
			//remove all matching-fixes
			fixedMatches = Arrays.copyOf(fixCopy,n);
			fixedMatches[confVert]=true;
			currentMatching[confVert]=v;
			//recurse only if the new choice for v fixes conflict for currentPos
			if (!areConflict(currentPos,confVert)){
				if (resolveMatching()) return true;
			}
		}
		return false;
	}
	 
	private Conflict getConflict(int maxPos){
		Conflict result = new Conflict(ConflictType.None);
		for (int i=0;i<=maxPos;i++){
			for (int j=0;j<i;j++){
				//System.out.println(i+" "+j);
				if (areConflict(i,j)){
					if (fixedMatches[i]&&fixedMatches[j]) return new Conflict(ConflictType.Unresolvable);
					if (result.type==ConflictType.None) result = new Conflict(i, j, ConflictType.Edge);					 						
				}
				else if (currentMatching[i]==currentMatching[j]){
					if (fixedMatches[i]&&fixedMatches[j]) return new Conflict(ConflictType.Unresolvable);
					if (result.type==ConflictType.None) result = new Conflict(i, j, ConflictType.Label);
				}
			}	
		}
		//no conflicts found, will return empty conflict
		return result;
	}
	
	private Map<String,String> getMatchingMap(){
		Map<String,String> result = new HashMap<String,String>();
		for (int i=0;i<n;i++){
			result.put(g.getVertexString(i),h.getVertexString(currentMatching[i]));
		}
		return result;
	}

}
