package algorithms;

import graph.AdjacencyMatrixGraph;
import graph.ListColoredAdjacencyMatrixGraph;
import importer.Results.Instance;

import java.util.*;

public class SearchTreeSolver extends Solver {
	final double timeBound = 60000*30;
	double startTime;
	private MatchingType type;
	private final boolean verbose = false;
	private Instance instance;
	
	public SearchTreeSolver(ListColoredAdjacencyMatrixGraph g, AdjacencyMatrixGraph h, MatchingType m, Instance instance){
		super(g,h);
		type = m;
		this.instance = instance;
	}
	
	@Override
	public boolean solve() {
		startTime = System.currentTimeMillis();
		ReductionRules r = new ReductionRules(this);
		r.reduceInstance();
		return findMatch(0);
	}
	
	/**
	 * Recursive search tree method that finds a matching of the graphs 
	 *
	 * @param  i  the first non-matched vertex
	 */
	private boolean findMatch(int i){		
		if (verbose) System.out.println("i: "+i);
		if (System.currentTimeMillis()-startTime>timeBound) {
			instance.terminate();
			return false;
		}
		if (i==g.vertexNum()) {
			if (verbose) System.out.println("Match found");
			instance.addMatching(getMatchingMap());
			return true;
		}
		for (int v :g.getList(i)){
			
			if (verbose) System.out.println(i+" matched to "+v);
			if (isMatchable(i, v)){//vertex i in g can be matched to v in h
				if (branch(i,v)) return true;
			}
		}
		return false;		
	}
	
	private boolean isMatchable(int i, int v){
		for (int j=0;j<i;j++){													//iterate over all vertices that are matched so far			
			if (this.type==MatchingType.Homomorph){								//Homomorphism test: if {i,j} in G, then matches have to be adjacent
				if (g.areAdjacent(i, j)){										
					int w = g.getList(j).iterator().next(); 					//j should only have one match left (since j<i)
					if (!h.areAdjacent(w, v)) return false; 					
				}
			}
			if (this.type==MatchingType.Isomorph){				
				int w = g.getList(j).iterator().next(); 					//j should only have one match left (since j<i)
				if (g.areAdjacent(i,j) != h.areAdjacent(w, v)) return false; //then their matching verts in h must be also adjacent								
			}
		}
		return true;
	}
		
	private boolean branch(int i, int v){
		Set<Integer> newList = new HashSet<Integer>();
		newList.add(v);
		Set<Integer> oldList = g.setList(i, newList);		//i is matched to v, old List is stored
		Set<Integer> matchesOfV = new HashSet<Integer>();
		for (int j=i+1;j<g.vertexNum();j++){
			if (g.getList(j).contains(v)){
				matchesOfV.add(j);
				g.getList(j).remove(v);
			}
		}
		boolean result = findMatch(i+1);		
		g.setList(i, oldList);						//restore old status
		for (int vmatch: matchesOfV) g.getList(vmatch).add(v);
		return result;
	}
		
	private Map<String,String> getMatchingMap(){
		Map<String,String> result = new HashMap<String,String>();
		for (int i=0;i<g.vertexNum();i++){
			int j = g.getList(i).iterator().next();
			result.put(g.getVertexString(i),h.getVertexString(j));
		}
		return result;
	}
	
}
