package main;

import graph.*;
import algorithms.*;
import importer.*;
import importer.Results.Instance;

import java.io.IOException;
import java.util.*;

/**
 * A Java Program for solving bounded-list network querying problems. 
 * The input is specified as follows<br>
 * A list of complexes, that is, subsets of vertices of G who should be queried in H<br>
 * A graph G (that contains the query)<br>
 * A graph H (the pattern)<br>
 * A list of similarities between vertices in G and H<br>
 * A similarity threshold t, for considering vertices in G and H similar.<br>
 * <br>
 * For each complex C in the complex list, the program finds (if it exists) a subnetwork of H 
 * that is similar to the subnetwork G[C] of G that is induced by C. 
 * Herein, similarity has two "components":
 * First, we want to map each vertex in C only to a 
 * similar vertex in H (similarity is specified by 
 * the similarity file and the similarity threshold p).
 * Second, the topology of the graph that is induced by 
 * the matching should be similar to the topology of G[C].<br><br>
 * 
 * The following options are available for considering subnetworks as similar:<br>
 * Injective Homomorphism - asks for a mapping from the vertices in G[C] to the vertices in H that is an injective homomorphism
 * . That is, for two vertices that are adjacent in G[C], the matches have to be adjacent in H.<br>
 * Subgraph Isomorphism - asks for a mapping from the vertices in G[C] to the vertices in H<br> 
 * that is an isomorphism. That is, it is a mapping from C to a vertex set S (from H), that is an isomorphism from G[C] to H[S].
 * 
 * You should have received the following example data with this program.<br>
 * 
 * - two protein interaction networks, one with yeast proteins, one with human proteins<br>
 * - a list of human protein complexes<br>
 * - a file that contains sequence similarities for yeast and human proteins
 * <br><br>
 * When using this data, please cite the following publication:<br>
 * 
 * Sharon Bruckner, Falk Hüffner, Richard M. Karp, 
 * Ron Shamir, Roded Sharan: Topology-Free Querying of Protein Interaction Networks. RECOMB 2009: 74-89
 * 
 * 
 * <br><br>Source: http://ftp.akt.tu-berlin.de/BLM<br>
 * For questions, remarks etc., please send an e-mail to Christian Komusiewicz (christian.komusiewicz@tu-berlin.de)
 * 
 * @author Christian Komusiewicz
 * @version 0.1
 * 
 * 
 */

public class BLM {	
	static ListColoredAdjacencyMatrixGraph g,h;
	static ParseGraph parser;
	static HashMap<String,List<String>> matches;
	static Solver.MatchingType matchType = Solver.MatchingType.Homomorph;
	static Solver.AlgorithmType algoType  = Solver.AlgorithmType.IterativeSearchTree;  // e.g. Solver.AlgorithmType.SearchTree;
	static String complexPath, gPath, hPath, matchPath, outputPath;
	static int minSize = 3;
	static double matchThreshold = 1E-7;
	static long startTime,taskTime,overallTime=0;
	static final boolean verbose = false;
	
	/**
	 * Main method of the Bounded List Morphism Program, 
	 * supplies different algorithms for solving the Bounded List Morphism problem for two graphs. 
	 * @param args - A string array that contains paths to the input files and further options. 
	 * It should contain the following, in the same order:<br>
	 * 1) A Path to the complex file. Each line of the file should consist of a complex in the following format: <br>
	 * Complex Name \t Protein1 \space Protein2<br>
	 * 2) A Path to the graph that contains the query, each line representing an edge: Protein_1 \space Protein_2<br>
	 * 3) A Path to the host, in the same format as above<br>
	 * 4) A Path to the similarity file. Each line specifies a similarity, given by BLAST E-values between two proteins:
	 * Protein_1 \space Protein_2 \space E-value
	 * 5) Sets the output path for the reported matches<br>
	 * The following arguments are optional, each has a default value<br>
	 * 6) Sets the E-value threshold for orthology; default value is 1E-7
	 * 7) Sets the mapping option; default is Injective Homomorphism; "2" is Subgraph Isomorphism	 
	 * 8) Sets the minimum size for reporting matches; default minimum size is 3
	 * 9) Sets the algorithm option; default is the iterative algorithm; "2" is the search tree algorithm
	 * 
	 */
	public static void main(String[] args) throws IOException{
		
		
		if (args.length>=5){ 
			
			/* 
			 * Setting input parameters 
			 */
			complexPath = args[0];							//Path of complex file
			gPath = args[1];								//Path of graph that contains complex/query
			hPath = args[2];								//Path of graph that is queried
			matchPath = args[3];							//Path of Matching scores between graph vertices
			outputPath = args[4];
			
			if (args.length >5)								//Orthology threshold; default = 1E-7
				matchThreshold = Double.valueOf(args[5]);		
			
			if (args.length >6){
				if (Integer.parseInt(args[6])==2) 				//Mapping option 
					matchType = Solver.MatchingType.Isomorph;	//Default = Homomorphism, 2 = Isomorphism
			}			
			
			if (args.length>7)								//MinSize; Default = 3
					minSize = Integer.parseInt(args[7]);
			
	
			if (args.length>8){								//Algorithm option
				if (Integer.parseInt(args[8])==2){			//Default = iterative, 2=Search Tree
					algoType = Solver.AlgorithmType.SearchTree;
				}
			}
			
			/*
			 * Initializing the output
			 */
			System.out.println("Creating "+outputPath+matchType.toString()+"/");
			Results result = new Results(outputPath+matchType.toString()+"/");
			
			/*
			 * Initializing parser Object + parsing of complexes
			 */
			parser = new ParseGraph();
			List<Complex> complexList = parser.parseComplexList(complexPath); 
			
			/*
			 * Main algorithm loop, processes each complex of the list
			 */
			for (Complex c:complexList){ 												//processing each complex in the list
				Set<String> complex = c.prots;
				if (verbose) System.out.println("Complex: "+complex);
				
				matches = parser.parseMatchList(complex, matchPath, matchThreshold);	//get the matches above threshold
				Set<String> subcomplex= new HashSet<String>();
				for (String prot: matches.keySet()){									//only add matched proteins from complex						
					if (!matches.get(prot).isEmpty()) subcomplex.add(prot); 
				}
				matches.keySet().retainAll(subcomplex);									//remove unmatched proteins from 
																						//candidate set
				
				if (verbose) 
					System.out.println("Subcomplex: " + subcomplex);
				
				if (!subcomplex.isEmpty()){
					g = parser.parseInducedGraph(subcomplex, gPath);					//parse graph induced by complex nodes					
					Set<String> homoVertices = new HashSet<String> ();					//obtain the vertex set of other graph
					for (String prot: matches.keySet()){
						for (String protB : matches.get(prot)){
							homoVertices.add(protB);
						}
					}
					if (verbose) System.out.println("Set of homolog proteins" + homoVertices);
					Instance instance = result.addInstance(c.name,complex,subcomplex,homoVertices);
					
					h = parser.parseInducedGraph(homoVertices, hPath);					//get graph induced by homologs								//parse graph induced by homolog nodes
					g.addColorList(parser.parseColors(matches, g, h));					//add color-lists to g
					if (verbose) System.out.println(g.toString());
					if (verbose) System.out.println(h.toString());
					
					/*
					 * Create Solver object and solve the instance, record
					 * running time and add to output
					 */
					Solver solver = Solver.solverFactory(g,h,matchType,instance,algoType);//Create Solver instance and solve problem
					startTime = System.currentTimeMillis();
					solver.solve();
					taskTime = System.currentTimeMillis() - startTime;
					instance.addRunningTime(taskTime);					
					overallTime += taskTime;
					
				}			
			}
			System.out.println("Overall running time:" +overallTime);
			result.addRunningTime(overallTime);
			result.writeMatches(minSize);
			result.writeStatistics();
		}
		else System.out.println("Usage: java BLM " + 
				"complexFile " +
				"queryFile " +
				"hostFile " +
				"similarityFile " +
				"outputPath " + 
				"orthologyThreshold (optional) " +
				"mappingType (optional) " +
				"minSize (optional) " +
				"algoType (optional) ");
	}	
}
